/* 
	dlp_th1.c
	Version 0.2
	R.G. Bankras

	Based on source code by Eric Nutt published on dlpdesign.com
*/

#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include "dlp_th1.h"
#include "ftd2xx.h"

typedef struct {
	// USB device info
	DWORD vid;
	DWORD pid;
	char *desc;
	char sn[64];
	FT_HANDLE  ftHandle;

	// Data
	unsigned char data;
	unsigned char dowcrc;
	
	// Measurements
	double degc;
	double hum;
	double dewp;
	
	// status flags
	char isAddedVidPid;
	char isOpen;
	char isCollecting;
} _dlp_th1;

int dlp_init(_dlp_th1 *dlp) {
	FT_STATUS ftStatus;
	FT_HANDLE ftHandle;
	int numDevs, Flags, ID, Type, LocId;
	char SerialNumber[16];
	char Description[64];

	// initialize structure members
	dlp->vid = DLPTH1VID;
	dlp->pid = DLPTH1PID;
	dlp->desc = DLPTH1DESC;

	dlp->data = 0x00;
	
	dlp->isAddedVidPid = FALSE;
	dlp->isOpen = FALSE;
	dlp->isCollecting = FALSE;
	
	// add DLP-TH1 VID/PID info to table
	ftStatus = FT_SetVIDPID(dlp->vid, dlp->pid);
	if (ftStatus != FT_OK) {
		if (DEBUG) { printf("X dlp_init: FT_SetVIDPID failed!\n"); }
		return FAILURE;
	} else {
		if (DEBUG) { printf("O dlp_init: FT_SetVIDPID success.\n"); }
		dlp->isAddedVidPid = TRUE;
	}
	
	// read number of D2XX devices connected to the system
	ftStatus = FT_CreateDeviceInfoList((DWORD *)&numDevs);
	if (ftStatus != FT_OK) {
		if (DEBUG) { printf("X dlp_init: FT_CreateDeviceInfoList failed!\n"); }
		return FAILURE;
	} else {
		if (DEBUG) { printf("O dlp_init: FT_CreateDeviceInfoList success: %i D2XX devices connected.\n", numDevs); }
		dlp->isAddedVidPid = TRUE;
	}	

	// read device info of first device
	ftStatus = FT_GetDeviceInfoDetail(0, (DWORD *)&Flags, (DWORD *)&Type, (DWORD *)&ID, (DWORD *)&LocId, SerialNumber, Description, &ftHandle);
	if (ftStatus != FT_OK) {
		if (DEBUG) { printf("X dlp_init: FT_GetDeviceInfoDetail failed!\n"); }
		return FAILURE;		
	} else {
		if (DEBUG) {
			printf("O dlp_init: FT_GetDeviceInfoDetail success:\n");
			printf("		Flags=0x%x\n",Flags);
			printf("		Type=0x%x\n",Type);
			printf("		ID=0x%x\n",ID);
			printf("		LocId=0x%x\n",LocId);
			printf("		SerialNumber=%s\n",SerialNumber);
			printf("		Description=%s\n",Description);
		}
	}

	return SUCCESS;
}

int dlp_open(_dlp_th1 *dlp) {
	FT_STATUS ftStatus;
	
	// check if device is already open
	if (dlp->isOpen) {
		if (DEBUG) { printf("X dlp_open: device is already open!\n"); }
		return SUCCESS;
	}
	
	if (!dlp->isAddedVidPid) { 
		if (DEBUG) { printf("X dlp_open: structure is not initialized!\n"); }
		return FAILURE; 
	}
	
	// Try to open last found DLP-TH1
	ftStatus = FT_Open(0, &(dlp->ftHandle));
	if (ftStatus != FT_OK) { 
		if (DEBUG) { printf("X dlp_open: FT_Open failed to open DLP-TH1 device!\n"); }
		return FAILURE; 
	}

	// set timeouts
	ftStatus = FT_SetTimeouts(dlp->ftHandle, 5000, 1000);		// read timeout 5 sec, write timeout of 1 sec
	if (ftStatus != FT_OK) {
		if (DEBUG) { printf("X dlp_open: FT_SetTimeouts failed!\n"); }
		FT_Close(dlp->ftHandle);
		return FAILURE;
	}
			
	// enter bit-bang mode
	ftStatus = FT_SetBitMode(dlp->ftHandle, 0xFF, DLPTH1BITMODE);
	if (ftStatus != FT_OK) {
		if (DEBUG) { printf("X dlp_open: Could not enable bit-bang mode!\n"); }
		FT_Close(dlp->ftHandle);
		return FAILURE;
	}

	dlp->isOpen = TRUE;
	if (DEBUG) { printf("O dlp_open: Device opened and bit-bang mode enabled.\n"); }

	return SUCCESS;
}

int dlp_close(_dlp_th1 *dlp) {
	FT_STATUS ftStatus;
	
	// check if device is open
	if (!dlp->isOpen) {
		if (DEBUG) { printf("X dlp_close: Device is not open!\n"); }
		return FAILURE;
	} else {
		if (DEBUG) { printf("O dlp_close: Device is open and will be closed.\n"); }
	}
	
	// close the device
	ftStatus = FT_Close(dlp->ftHandle);
	if (ftStatus != FT_OK) {
		if (DEBUG) { printf("X dlp_close: Failed to close device.\n"); }
		return FAILURE;
	} else {
		if (DEBUG) { printf("O dlp_close: Device closed.\n"); }
		dlp->isOpen = FALSE;
	}
	
	return SUCCESS;
}

int dlp_set_data_direction(_dlp_th1 *dlp, int dir) {
	FT_STATUS ftStatus;
	
	// check if device is open
	if (!dlp->isOpen) {
		if (DEBUG) { printf("X dlp_set_data_direction: Device is not open!\n"); }
		return FAILURE;
	} else {
		if (DEBUG) { 
			printf("O dlp_set_data_direction: Direction of bit-bang data bit will be set to ");
			if (dir == DLPTH1INPUT) {
				printf("INPUT.\n");
			} else {
				printf("OUTPUT.\n");
			}
		}
	}		

	if (dir == DLPTH1INPUT) {
		ftStatus = FT_SetBitMode(dlp->ftHandle, 0xFD, DLPTH1BITMODE);		// bit 1 set to low: input
	} else {
		ftStatus = FT_SetBitMode(dlp->ftHandle, 0xFF, DLPTH1BITMODE);		// bit 1 set to high: output
	}
	if (ftStatus != FT_OK) {
		if (DEBUG) { printf("X dlp_set_data_direction: Change bit-bang data direction failed!\n"); }
		return FAILURE;
	} else {
		if (DEBUG) { printf("O dlp_set_data_direction: Bit-bang data direction changed.\n"); }
	}		
	
	return SUCCESS;
}

int dlp_set_clear_bit(_dlp_th1 *dlp, int hilow, int bitpos) {
	FT_STATUS ftStatus;
	DWORD ret_bytes;
	unsigned char mask;	
	
	// check for valid arguments
	if (dlp == NULL || bitpos < 0 || bitpos > 7) {
		if (DEBUG) { printf("X dlp_set_clear_bit: Invalid arguments!\n"); }
		return FAILURE;
	}
	
	// setup a mask with bit set in bitpos
	mask = 0x01;
	mask <<= bitpos;
	
	// invert mask for clearing
	if (hilow == 0) mask = ~mask;
	
	// set or clear bit in dlp->data using mask
	dlp->data = (hilow == 0) ? dlp->data & mask : dlp->data | mask;
	
	ftStatus = FT_Write(dlp->ftHandle, &dlp->data, 1, &ret_bytes);
	if (ftStatus != FT_OK || ret_bytes != 1) {
		if (DEBUG) { printf("X dlp_set_clear_bit: Write to dlp-th1 failed!\n"); }
		return FAILURE;
	} else {
		if (DEBUG) { printf("O dlp_set_clear_bit: Write to dlp-th1 success: data 0x%02x\n", dlp->data); }
	}	
	return SUCCESS;
}

int dlp_transmission_start(_dlp_th1 *dlp) {
	// check if device is open
	if (!dlp->isOpen) {
		if (DEBUG) { printf("X dlp_transmission_start: Device is not open!\n"); }
		return FAILURE;
	} else {
		if (DEBUG) { printf("O dlp_transmission_start: Will send transmission start signal.\n"); }
	}	

	dlp_set_data_direction(dlp, DLPTH1OUTPUT);		// direction = output	
	dlp_set_clear_bit(dlp, SET, DLPTH1DATAPOS);		// set DATA
	dlp_set_clear_bit(dlp, SET, DLPTH1SCKPOS);		// set SCK
	dlp_set_clear_bit(dlp, CLEAR, DLPTH1DATAPOS);		// clear DATA
	dlp_set_clear_bit(dlp, CLEAR, DLPTH1SCKPOS);		// clear SCK
	dlp_set_clear_bit(dlp, SET, DLPTH1SCKPOS);		// set SCK
	dlp_set_clear_bit(dlp, SET, DLPTH1DATAPOS);		// set DATA
	dlp_set_clear_bit(dlp, CLEAR, DLPTH1SCKPOS);		// clear SCK
	dlp_set_clear_bit(dlp, CLEAR, DLPTH1DATAPOS);		// clear DATA

	if (DEBUG) { printf("O dlp_transmission_start: Completed sending transmission start signal.\n"); }

	return SUCCESS;
}

int dlp_write_byte(_dlp_th1 *dlp, int data) {
	int count, bitvalue;
	
	if (!dlp->isOpen) {
		if (DEBUG) { printf("X dlp_write_byte: Device is not open!\n"); }
		return FAILURE;
	}
	
	for (count=8; count>0; count--) {
		bitvalue = data & 0x80;		// select bit 7 as value to write
		
		dlp_set_clear_bit(dlp, bitvalue, DLPTH1DATAPOS);	
		dlp_set_clear_bit(dlp, SET, DLPTH1SCKPOS);
		dlp_set_clear_bit(dlp, CLEAR, DLPTH1SCKPOS);
		
		data <<= 1;			// shift next bit in
	}

	// send ACK
	dlp_set_clear_bit(dlp, CLEAR, DLPTH1DATAPOS);
	dlp_set_clear_bit(dlp, SET, DLPTH1SCKPOS);
	dlp_set_clear_bit(dlp, CLEAR, DLPTH1SCKPOS);

	if (DEBUG) { printf("O dlp_write_byte: Completed writing byte.\n"); }
	return SUCCESS;
}

int dlp_read_byte(_dlp_th1 *dlp, int lastbyte) {
	FT_STATUS ftStatus;
	int count, bitvalue, bytevalue;

	if (!dlp->isOpen) {
		if (DEBUG) { printf("X dlp_read_byte: Device is not open!\n"); }
		return FAILURE;
	}
	
	// init local vars
	bitvalue = 0;
	bytevalue = 0;

	dlp_set_data_direction(dlp, DLPTH1INPUT);		// direction = input	
	
	// read in the byte
	for (count=8; count>0; count--) {
		dlp_set_clear_bit(dlp, SET, DLPTH1SCKPOS);
		
		// read the data bit state
		ftStatus = FT_GetBitMode(dlp->ftHandle, (unsigned char *)&bitvalue);
		if (ftStatus != FT_OK) {
			if (DEBUG) { printf("X dlp_read_byte: FT_GetBitMode failed!\n"); }
			return FAILURE;
		}

		dlp_set_clear_bit(dlp, CLEAR, DLPTH1SCKPOS);

		bitvalue &= 0x02;				// select the data bit
		bytevalue <<= 1;				// shift left by 1 bit
		if (bitvalue) bytevalue |= 1;			// add new data bit to byte
	}
	
	// send ACK
	dlp_set_data_direction(dlp, DLPTH1OUTPUT);		// direction = output	
	if (lastbyte == 0) dlp_set_clear_bit(dlp, CLEAR, DLPTH1DATAPOS);
	dlp_set_clear_bit(dlp, SET, DLPTH1SCKPOS);
	dlp_set_clear_bit(dlp, CLEAR, DLPTH1SCKPOS);
	
	return bytevalue;
}

int dlp_configure(_dlp_th1 *dlp) {
	int readbyte;

	if (!dlp->isOpen) {
		if (DEBUG) { printf("X dlp_configure: Device is not open!\n"); }
		return FAILURE;
	}

	dlp_transmission_start(dlp);
	dlp_write_byte(dlp, DLPTH1CMDWRITESTATUS);
	dlp_write_byte(dlp, 0x00);

	dlp_transmission_start(dlp);
	dlp_write_byte(dlp, DLPTH1CMDREADSTATUS);
	readbyte = dlp_read_byte(dlp, 0);
	if (DEBUG) { printf("O dlp_configure: Read status register (%i).\n", readbyte); }
	readbyte = dlp_read_byte(dlp, 1);
	if (DEBUG) { printf("O dlp_configure: Read status register CRC (%i).\n", readbyte); }

	if (DEBUG) { printf("O dlp_configure: Completed writing status register.\n"); }
	return SUCCESS;
}

int dlp_calcrc(_dlp_th1 *dlp, unsigned char x) {

   static unsigned char dscrc_table[256]= {
	0, 49, 98, 83, 196, 245,166, 151,185, 136,219, 234,125, 76, 31, 46, 67, 114,33, 16,
	135, 182,229, 212,250, 203,152, 169,62, 15, 92, 109,134, 183,228, 213,66, 115,32, 17,
	63, 14, 93, 108,251, 202,153, 168,197, 244,167, 150,1, 48, 99, 82, 124, 77, 30, 47,
	184, 137,218, 235,61, 12, 95, 110,249, 200,155, 170,132, 181,230, 215,64, 113,34, 19,
	126, 79, 28, 45, 186, 139,216, 233,199, 246,165, 148,3, 50, 97, 80, 187, 138,217, 232,
	127, 78, 29, 44, 2, 51, 96, 81, 198, 247,164, 149,248, 201,154, 171,60, 13, 94, 111,
	65, 112,35, 18, 133, 180,231, 214,122, 75, 24, 41, 190, 143,220, 237,195, 242,161, 144,
	7, 54, 101, 84, 57, 8, 91, 106,253, 204,159, 174,128, 177,226, 211,68, 117,38, 23,
	252, 205,158, 175,56, 9, 90, 107,69, 116,39, 22, 129, 176,227, 210,191, 142,221, 236,
	123, 74, 25, 40, 6, 55, 100, 85, 194, 243,160, 145,71, 118,37, 20, 131, 178,225, 208,
	254, 207,156, 173,58, 11, 88, 105,4, 53, 102, 87, 192, 241,162, 147,189, 140,223, 238,
	121, 72, 27, 42, 193, 240,163, 146,5, 52, 103, 86, 120, 73, 26, 43, 188, 141,222, 239,
	130, 179,224, 209,70, 119,36, 21, 59, 10, 89, 104,255, 206,157, 172
   };

   dlp->dowcrc = dscrc_table[dlp->dowcrc ^ x];	// ^ = XOR

   return SUCCESS;
}

int dlp_read_temp(_dlp_th1 *dlp) {
	FT_STATUS ftStatus;
	int busy, i, measurement;
	unsigned char msb, lsb, crc, ret_crc;
	
	if (!dlp->isOpen) {
		if (DEBUG) { printf("X dlp_read_temp: Device is not open!\n"); }
		return FAILURE;
	}

	dlp_transmission_start(dlp);	
	dlp_write_byte(dlp, DLPTH1CMDMEASTEMP);
	dlp_set_data_direction(dlp, DLPTH1INPUT);	
	
	// wait for conversion to finish
	busy = 1;
	while (busy) {
		ftStatus = FT_GetBitMode(dlp->ftHandle, &dlp->data);
		if (ftStatus != FT_OK) {
			if (DEBUG) { printf("X dlp_read_temp: FT_GetBitMode failed!\n"); }
			return FAILURE;
		}
		busy = dlp->data & 0x02;
		// sleep for 10 ms
		usleep(10000);
	}
	
	// read temp data
	msb = dlp_read_byte(dlp, 0);
	lsb = dlp_read_byte(dlp, 0);
	crc = dlp_read_byte(dlp, 1);

	// process CRC for the command byte
	dlp->dowcrc = 0;
	dlp_calcrc(dlp, 3);
	dlp_calcrc(dlp, msb);
	dlp_calcrc(dlp, lsb);
	
	// reverse bit order of the returned checksum
	ret_crc = crc;
	crc = 0;
	for (i=0; i<8; i++) {
		crc <<= 1;
		if (ret_crc & 0x01) crc |= 0x01;
		ret_crc >>= 1;
	}
	
	if (crc != dlp->dowcrc) {
		if (DEBUG) { printf("X dlp_read_temp: CRC of temperature reading failed!\n"); }
		dlp->degc = DLPTH1INVALID;
		return FAILURE;
	} else {
		if (DEBUG) { printf("O dlp_read_temp: CRC of temperature reading passed.\n"); }
	}


	measurement = lsb | (msb << 8);
	dlp->degc = (-40.10 + (0.01 * (double)measurement));	// Vdd = 5.0 V, 14-bit measurement

	if (DEBUG) { printf("X dlp_read_temp: temperature =  %.1f C.\n", dlp->degc); }

	return SUCCESS;
}

int dlp_read_hum(_dlp_th1 *dlp) {
	FT_STATUS ftStatus;
	int busy, i, measurement;
	unsigned char msb, lsb, crc, ret_crc;
	double rhlin, rhtrue, Tn, m, a, b;
	
	if (!dlp->isOpen) {
		if (DEBUG) { printf("X dlp_read_hum: Device is not open!\n"); }
		return FAILURE;
	}

	dlp_transmission_start(dlp);	
	dlp_write_byte(dlp, DLPTH1CMDMEASHUM);
	dlp_set_data_direction(dlp, DLPTH1INPUT);	
	
	// wait for conversion to finish
	busy = 1;
	while (busy) {
		ftStatus = FT_GetBitMode(dlp->ftHandle, &dlp->data);
		if (ftStatus != FT_OK) {
			if (DEBUG) { printf("X dlp_read_hum: FT_GetBitMode failed!\n"); }
			return FAILURE;
		}
		busy = dlp->data & 0x02;
		// sleep for 10 ms
		usleep(10000);
	}
	
	// read humidity data
	msb = dlp_read_byte(dlp, 0);
	lsb = dlp_read_byte(dlp, 0);
	crc = dlp_read_byte(dlp, 1);

	// process CRC for the command byte
	dlp->dowcrc = 0;
	dlp_calcrc(dlp, 5);
	dlp_calcrc(dlp, msb);
	dlp_calcrc(dlp, lsb);
	
	// reverse bit order of the returned checksum
	ret_crc = crc;
	crc = 0;
	for (i=0; i<8; i++) {
		crc <<= 1;
		if (ret_crc & 0x01) crc |= 0x01;
		ret_crc >>= 1;
	}
	
	if (crc != dlp->dowcrc) {
		if (DEBUG) { printf("X dlp_read_hum: CRC of humidity reading failed!\n"); }
		dlp->hum = DLPTH1INVALID;
		return FAILURE;
	} else {
		if (DEBUG) { printf("O dlp_read_hum: CRC of humidity reading passed.\n"); }
	}

	measurement = lsb | (msb << 8);
	rhlin = -2.0468 + (0.0367*(double)measurement) + (-1.5955e-6*(double)measurement*(double)measurement);
	rhtrue = ((dlp->degc - 25.0) * (0.01 + .00008 * measurement)) + rhlin;
	dlp->hum = rhtrue;

	if (dlp->degc > 0) {
		Tn = 243.12;
		m = 17.62;
	} else {
		Tn = 272.62;
		m = 22.46;
	}

	a = m * dlp->degc / (Tn + dlp->degc);
	b = log(dlp->hum / 100);
	dlp->dewp = Tn * (a + b) / (m - a - b);

	if (DEBUG) {
		printf("O dlp_read_hum: humidity =  %.1f %%.\n", dlp->hum);
		printf("O dlp_read_hum: dewpoint =  %.1f C.\n", dlp->dewp);
	}

	return SUCCESS;
}

int main() {
	// main dlp-th1 structure
	_dlp_th1 dlpth1;
	
	printf("\n-----------------------------------------------------------------------------------\n");

	// initialize th1 structure values
	if (dlp_init(&dlpth1) != SUCCESS) {
		if (DEBUG) { printf("X main: Function dlp_init failed!\n"); }
		return -1;
	} else {
		if (DEBUG) { printf("O main: Function dlp_init success.\n"); }
	}
	
	// attempt to find and open DLP-TH1
	if (dlp_open(&dlpth1) != SUCCESS) {
		if (DEBUG) { printf("X main: Function dlp_open failed!\n"); }
		return -1;
	} else {
		if (DEBUG) { printf("O main: Function dlp_open success.\n"); }
	}

	// set advanced functions of SHT11
	if (dlp_configure(&dlpth1) != SUCCESS) {
		if (DEBUG) { printf("X main: Function dlp_configure failed!\n"); }
		return -1;
	} else {
		if (DEBUG) { printf("O main: Function dlp_configure success.\n"); }
	}

	// read temperature
	if (dlp_read_temp(&dlpth1) != SUCCESS) {
		if (DEBUG) { printf("X main: Function dlp_read_temp failed!\n"); }
		return -1;
	} else {
		if (DEBUG) { printf("O main: Function dlp_read_temp success.\n"); }
	}

	// read humidity
	if (dlp_read_hum(&dlpth1) != SUCCESS) {
		if (DEBUG) { printf("X main: Function dlp_read_hum failed!\n"); }
		return -1;
	} else {
		if (DEBUG) { printf("O main: Function dlp_read_hum success.\n"); }
	}

	printf("Temperature: %.1f C\n", dlpth1.degc);
	printf("Humidity:    %.1f %%\n", dlpth1.hum);
	printf("Dewpoint:    %.1f C\n", dlpth1.dewp);

	// close the DLP-TH1 and exit
	if (dlp_close(&dlpth1) != SUCCESS) {
		if (DEBUG) { printf("X main: Function dlp_close failed!\n"); }
		return -1;
	} else {
		if (DEBUG) { printf("O main: Function dlp_close success.\n"); }
	}

	printf("-----------------------------------------------------------------------------------\n\n");
	
	return 0;
}

